/************************************************************************
   NOPPP.C  - M. Covington 1997, 1998, 1999
   Software for the "No-Parts Pic Programmer"
   Be sure to compile for 8088 (not 286 or 386) for maximum portability.
   
   Modified by Phil Rice 2000

   Ported to LINUX December 2004
   
   Command line arguments:  None.
************************************************************************

PP002.c		13/7/2004

   Modified by Phil Rice in an attempt to handle more devices
   
PP003.c		27/11/04	10:54pm

   Even more mods. Compiles OK on DOS QC, but hasn't been tested

pp003.c		6/12/04		9:54am

   Began porting to LINUX after discovering how to split the
   Process Scheduling Atom.
   
pp004.c		16/12/04	3:15pm

   used ioperm() to get permission to access I/O instead of iopl(3)
   Inlined some low level IO functions
   Fixed editing of "Load File" name
   
pp005.c		11/1/05		3:00pm

   crlf was a Global variable. Now local to readall()

   Rearranged IO Macros.

   Cleaned up getchoice().

   Added note re D4 to Interface info.

   Modified progcycle_12c5 to return status of location just blasted.

   Cleaned up main() (removed dead wood).

   Modified repair675() because it didn't allow line editing of the
   osccal and bandgap values. (Needed to "stty stdin").
   
   Modified selectdevice() so it zaps the hex file name and PIC type.
   
   Modified repair675() and erase675() so they both correctly set the
   osccal and bandgap values (and they didn't properly erase the pic).
   
   Changed 16F628 erase function to be the same as the 16F84 cos it works
   
************************************************************************/
#define LINUX

#ifndef LINUX
#define	DOS
#endif

#define	FALSE	0
#define	TRUE	1
#define	LOW	0
#define	HIGH	1
#define	EARLY	0
#define	LATE	1

typedef	unsigned int	word;
typedef	unsigned char	byte, bit;

#include	<stdio.h>
#include	"pp005.h"

#ifdef DOS
//#include	<pc.h>
#include	<dos.h>
#include	<bios.h>
#include	<conio.h>
#endif

#ifdef LINUX
#include <time.h>      
#include <sched.h>      
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/io.h>      
#endif

#include	<malloc.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	<time.h>

#ifdef LINUX

#define OUTPORT outb(DATA, PORT)

#define HIGH_P	fooey.sched_priority=10;\
		sched_setscheduler( 0,SCHED_FIFO, &fooey);
		
#define LOW_P   fooey.sched_priority=0;\
		sched_setscheduler( 0,SCHED_OTHER, &fooey);
		
#define CLAIM_PORT	ioperm( PORT, 3, 1 );

#define UNCLAIM_PORT	ioperm( PORT, 3, 0 );

#endif

#ifdef DOS
#define OUTPORT outp(PORT, DATA)
#define HIGH_P  
#define LOW_P  
#define CLAIM_PORT	ioperm( PORT, 3, 1 );
#define UNCLAIM_PORT	ioperm( PORT, 3, 0 );
#endif

/************************************************************************
   Macros for some simpler I/O actions
   These were functions - now inlined
   Remember, there is an inverter in the programmer
************************************************************************/

#define DATADOWN	{DATA |=  0x01;	OUTPORT; MyDelay(20);}
#define DATAUP		{DATA &= ~0x01;	OUTPORT; MyDelay(20);}
#define CLOCKDOWN	{DATA |=  0x02;	OUTPORT; MyDelay(20);}
#define CLOCKUP		{DATA &= ~0x02;	OUTPORT; MyDelay(20);}
#define VCCON		{DATA |=  0x04; OUTPORT; MyDelay(20000); MyDelay(20000);}
#define VCCOFF		{DATA &= ~0x04; OUTPORT; MyDelay(20000); MyDelay(20000);}

// Turn off both power settings
#define VPPOFF		{DATA &= ~0x18; OUTPORT; MyDelay(20000); MyDelay(20000);}

/************************************************************************
   GLOBAL STATUS VARIABLES
************************************************************************/

int	LPT  = 0;		// which LPT port we're using (1..3)
int	PORT = 0;		// its port address
byte	DATA = 0x0F;		// And what it is outputting. Why $0F??

#define PIC12C508   1
#define PIC12C508A  2
#define PIC12C509   3
#define PIC12C509A  4
#define PIC16C83    5
#define PIC16C84    6
#define PIC16F83    7
#define PIC16F84    8
#define PIC16F627   9
#define PIC16F628  10
#define PIC16F648  11
#define PIC12F675  12

int     DEVICE = 0;		// which PIC we're programming

#define	PROGRAM	1
#define	VERIFY	0		// desired action in main programming loop

char    FNAME[255];		// name of file currently loaded

char    CHOICE = 0;		// user's most recent menu choice

// PIC MEMORY MAP

// The PIC16F84 has four programmable memory areas
// (plus data RAM, which is not programmable).
// Config memory is only 1 byte, but is treated like the others.
// Dunno why I mention this?

#define PSIZEMAX   8192			// Max size of program memory
#define ISIZEMAX   4			// Max size of ID memory
#define CSIZEMAX   1			// Max size of Config memory
#define DSIZEMAX   256			// Max size of Data memory

word	PBASE       = 0x0000;		// Base address of each memory
word	IBASE       = 0x2000;		// Initialised by choice of PIC
word	CBASE       = 0x2007;
word	DBASE       = 0x2100;

word    PSIZE      = PSIZEMAX;		// Actual size, can be set lower
word    ISIZE      = ISIZEMAX;		// for particular CPUs
word    CSIZE      = CSIZEMAX;
word    DSIZE      = DSIZEMAX;

word    PMEM[PSIZEMAX];			// Arrays representing the memories
word    IMEM[ISIZEMAX];
word    CMEM[CSIZEMAX];
word    DMEM[DSIZEMAX];

word    PUSED      = 0;                // Number of valid words in array
word    CUSED      = 0;
word    IUSED      = 0;
word    DUSED      = 0;

word	PMASK      = 0x3fff;            // Which bits are used in each word
word    CMASK      = 0x001f;           //  (CMASK depends on processor)
word	IMASK      = 0x3fff;
word	DMASK      = 0x00ff;

word    DEFAULTCONFIG  =  0x1B;		// Initialization for config word

// PIC PROGRAMMING COMMANDS

byte LOADCONFIG        =  0;
byte LOADPROGRAM       =  2;
byte READPROGRAM       =  4;
byte INCREMENTADDRESS  =  6;
byte BEGINPROGRAMMING  =  8;
byte LOADDATA          =  3;
byte READDATA          =  5;
byte ERASEPROGRAM      =  9;
byte ERASEDATA         = 11;

byte ENDPROGRAMMING    = 14;

word bandgap           =  0;		// 12F675
word osccal            =  0;

void	(*erase_p)();			// Pointer to erase function
void	(*program_p)();			// Program/verify function
char	*pic_name = "         ";	// Name to display
int	Vpptime  = LATE;		// When to turn Vpp on
int	VPPlevel = LOW;			// How strong?

#ifdef LINUX
struct	sched_param	fooey;
struct	timespec	req, rem;
#endif

// *********************************************************************
// I/O UTILITIES
// *********************************************************************

#define KBUFSIZE 256

char buf[KBUFSIZE];          // keyboard input buffer

/**********************************************************************/

char* cleanctrl(char *s) {         // truncates string at first ctrl char,
	int i=0;                   // cleaning up ^M or ^J left behind by fgets
	while (s[i] >= ' ') i++;
	s[i] = 0;
	return s;
}

char* getstring() {                // read string

#ifdef LINUX
	system("stty -cbreak echo");
#endif

	fgets(buf,KBUFSIZE-1,stdin);

#ifdef LINUX
	system("stty cbreak -echo");
#endif

	cleanctrl(buf);
	return buf;
}

/**********************************************************************/

int getint()  { return atoi(getstring()); };     // read integer

/*   read menu choice
     prints prompt, then reads letters, uppercasing,
     until a letter that is in s is found             */

char getchoice(char *s)  {
  printf("\n\n  Your choice (%s): ",s);

  do {

#ifdef LINUX
    CHOICE = getchar();
#endif

#ifdef DOS
    CHOICE = getch();
#endif

    CHOICE = toupper(CHOICE);
    // toupper evals arg twice, so arg can't be getch()
  }
  while ((CHOICE == ',') || (strchr(s,CHOICE) == NULL));

  printf("  %c\n",CHOICE);
  return CHOICE;
}

/**********************************************************************/
 
int waitkey() {

	int	c;

#ifdef DOS
  while (kbhit()) { getch(); };
  puts("\n  Press space bar to continue...");
  getch();
  while (kbhit()) { getch(); };
#endif

#ifdef LINUX
	puts("\n  Press space bar to continue...");
	c = getchar();
#endif
	return c;
}

/**********************************************************************/

void clearscreen() {
#ifdef DOS
  _asm {
            mov ax,0f00h
            int 010h
            mov ah,0
            int 010h
  }
#endif

#ifdef LINUX
	system("clear");
#endif
}


/**********************************************************************/

void errmsg(char *s) {
  puts(s);
  waitkey();
}

// *********************************************************************
// TIMING
// *********************************************************************


void MyDelay(int k) {

#ifdef DOS

#define IODELAY (inp(0x61))   // allow time for timer to respond
  // Delay at least k microseconds, possibly a good bit more.
  // k must be less than 27304.
  // Minimum delay on a 25-MHz 386 is about 100 microseconds;
  // on a 133-MHz Pentium, about 18 microseconds.

  // Uses system timer 2.
  // When running in a DOS box under OS/2, set HW_TIMER ON in DOS settings.

  unsigned int w;
  unsigned char lo,hi;

  outp(0x61, (inp(0x61) & 0xFD) | 1);  // spkr off, tmr 2 gate on
  w = (unsigned int)(k*1.2);
  outp(0x43, 0xB0);                   // tmr 2 mode 0 2-byte load
  IODELAY;
  outp(0x42, (unsigned char)w);       // low byte
  IODELAY;
  outp(0x42, (unsigned char)(w>>8));  // high byte
  IODELAY;
  do {
    outp(0x43,0x80);          // latch timer count
    IODELAY;
    lo = inp(0x42);           // discard low byte
    IODELAY;
    hi = inp(0x42);           // get high byte
    IODELAY;
  }
  while ((hi & 0x80) == 0);    // wait for a 1 there, signifying rollover
  return;

#endif

#ifdef LINUX

	req.tv_sec  = 0;
	req.tv_nsec = 1000L * (long)k;
	nanosleep( &req, &rem);

#endif
}

// *********************************************************************
// PARALLEL PORT HARDWARE INTERFACE
// *********************************************************************//
/* Modified by Phil Rice to use following I/O lines

   Pin  2, D0 is data to be loaded into the PIC (inverted by programmer)
   Pin  3, D1 is clock to PIC (inverted by programmer)
   Pin  4, D2 controls Vcc. 1 = Vcc applied.
   Pin  5, D3 controls Vpp. 1 = Vpp applied.
   Pin  6, D4 controls High Power Vpp. 1 = High Power Vpp applied
   Pin 10, !Ack returns data from PIC (inverted by programmer, not by PC).
   Pins 18 to 25 are ground.
*/


/**********************************************************************/

word portaddr(int n) {                      // base address of port LPTn
#ifdef DOS
      return(*(word far *)(0x00000408L+2*n-2));
#endif

#ifdef LINUX
	switch(n) {
		case 1  : return(0x378);
		case 2  : return(0x278);
		case 3  : return(0x3bc);
		default : return(0x378);
	
	}
#endif
}

/**********************************************************************/
// Procedures to set and clear some data lines
/**********************************************************************/

void vppon() {

     if( VPPlevel == HIGH )
         DATA |=  0x18;  // Hi Power
     else
         DATA |=  0x08;  // Lo Power

     OUTPORT;
     MyDelay(20000);
     MyDelay(20000);
}

/**********************************************************************/

void allpinslow() {
  HIGH_P
  VPPOFF
  DATADOWN
  CLOCKDOWN
  VCCOFF
  LOW_P
}

/**********************************************************************/

bit  detecthardware() {  // true if datain follows dataup/down

int	access;

#ifdef LINUX
	access = CLAIM_PORT
#endif

#ifdef DOS
	access = 0; /* Don't bother checking in DOS */
#endif

	if( access != 0 )
		return(FALSE);
		
	DATAUP

	if(datain() != 1) {
		DATADOWN
		UNCLAIM_PORT
		return(FALSE);
	}

	DATADOWN

	if(datain() != 0) {
		UNCLAIM_PORT
		return(FALSE);
	}

   return(TRUE);
}

/********************************************************************
   PIC COMMUNICATION ROUTINES
********************************************************************/

bit datain() {

#ifdef DOS
	return(((~(byte)inp(PORT+1)) & 0x40) >> 6);
#endif

#ifdef LINUX
	return(((~(byte)inb(PORT+1)) & 0x40) >> 6);
#endif
}

/**********************************************************************/

void sendbit(bit b) {                  // Sends out 1 bit to PIC

  if(b) 
     DATAUP
  else
     DATADOWN

  MyDelay(10);
  CLOCKUP
  MyDelay(10);                         // tset1
  CLOCKDOWN                         // data is clocked into PIC on this edge
  MyDelay(10);                         // thld1
  DATADOWN                          // idle with data line low
  MyDelay(10);
}

/**********************************************************************/

bit recvbit() {                        // Receives a bit from PIC
  bit b;
  CLOCKUP
  MyDelay(10);                         // tdly3
  CLOCKDOWN                         // data is ready just before this
  MyDelay(10);
  b = datain();
  MyDelay(10);                         // thld1
  return b;
}

/**********************************************************************/

void sendcmd(byte b) {                 // Sends 6-bit command from bottom of b
  int i;
  MyDelay(220);                         // inc thld0
  for (i=6; i>0; i--) {
    sendbit((bit)(b & 1));
    b = b >> 1;
  }
  MyDelay(20);                         // tdly2
}

/**********************************************************************/

void senddata(word w) {                // Sends 14-bit word from bottom of w
  int i;
  MyDelay(220);                         // inc thld0
  sendbit(0);                          // one garbage bit
  for (i=14; i>0; i--) {
    sendbit((bit)(w & 1));             // 14 data bits
    w = w >> 1;
  }
  sendbit(0);                          // one garbage bit
  MyDelay(20);                         // tdly2
}

/**********************************************************************/

word recvdata() {			// Receives 14-bit word, lsb first
  int i;
  bit b;
  word w = 0;

  DATA &= ~0x01;			//  Just make sure data to the PIC is "0"
  OUTPORT;				//  so the OC inverter in the programmer
  MyDelay(220);				//  will let go of the Data I/O pin

              				//  inc thld0
  recvbit();				//  one garbage bit
  for (i=0; i<14; i++) {
    b = recvbit();
    w = w | ((word)b << i);		// 14 data bits
  }
  recvbit();				// another garbage bit;
  MyDelay(20);				// tdly2
  return w;
}

// *********************************************************************
// PIC PROGRAMMING ALGORITHMS
// *********************************************************************

void cleararrays () {                  // Mark the memory arrays as empty
  PUSED =
  IUSED =
  CUSED =
  DUSED = 0;
}

/**********************************************************************/

bit  stuffarray (word address,         // Stuff data into a memory array.
                 word array[],         // Returns true if successful.
                 word base,
                 word size,
                 word *used,
                 word data[],
                 int  count) {
  int i;
  if (address-base+count-1 > size) {
    printf("  Invalid address: %04XH\n",address+count-1);
    return 0;
  }
  for (i=0; i<count; i++) {
    array[i+address-base] = data[i];
    if (*used < address-base+count) *used = address-base+count;
  }
  return 1;
}

/**********************************************************************/
// PROGRAMMING PROCEDURES
/**********************************************************************/

void vppreset() {                      // reset PIC, apply Vpp

     HIGH_P

     VPPOFF
     MyDelay(200);
     DATADOWN
     CLOCKDOWN
     MyDelay(25000);

  if( Vpptime == EARLY ) {
     vppon();
     VCCON
  }else {
     VCCON
     vppon();
  }
  LOW_P
}

/**********************************************************************/

void progcycle(byte cmd, word arg) {   // send a command and an argument,
  sendcmd(cmd);                        // then program EPROM accordingly
  senddata(arg);
  sendcmd(BEGINPROGRAMMING);
  MyDelay(20000); // PIC errata sheet says 10ms is typical, not guaranteed
}

/**********************************************************************/

bit programall(
// Program a memory from appropriate array.
// Returns true if successful.
             int  mode,                // program or verify
             word mask,                // to throw away irrelevant bits
             byte writecommand,        // command to write this memory
             byte readcommand,         // command to read this memory,
             word array[],             // which memory array we're using
             word base,                // its base address
             word used) {              // number of valid words in array
  word i,w;
  switch(mode) {
    case PROGRAM: puts("  programming..."); break;
    case VERIFY:  puts("  verifying..."); break;
  }
  for (i=0; i<used; i++) {

    printf("  %04X\b\b\b\b\b\b",i+base);     // display addr and backspace
    fflush(stdout);
    
    HIGH_P
    if (mode==PROGRAM)
      progcycle(writecommand,(array[i] & mask));
    sendcmd(readcommand);
    w = (recvdata() & mask);
    if (w != (array[i] & mask)) {
      printf("  Failed at %04X: Expecting %04X, found %04X.\n",
                                         i+base, (array[i] & mask), w);
      LOW_P
      return 0;
    }
    sendcmd(INCREMENTADDRESS);
    LOW_P
  }
  return 1;
}


/* *********************************************************************
   HEX FILE LOADER

   This is for Intel INHX8M (8-bit merged) hex files only.
   These files use two bytes for each word (low, then high).
   All addresses are doubled, i.e., 0x2001 is encoded as 0x4002,
   so that addresses increment at the same rate as the byte count.

***********************************************************************/

bit validhexline(char *s) {            // Gross syntax and checksum check.
  byte cksum = 0;                      // For all HEX formats, not just 8M.
  int bytecount;
  int i, b;
  if (s[0] != ':') return(0);          // Initial colon
  sscanf(s+1,"%2x",&bytecount);
  if (bytecount > 32) return(0);       // Valid byte count
  cksum = bytecount;
  i = 3;
  bytecount = bytecount+3;
  while (bytecount>0) {
    bytecount--;
    sscanf(s+i,"%2x",&b);
    cksum = cksum+b;                   // Compute checksum
    i = i+2;
  }
  sscanf(s+i,"%2x",&b);
  cksum = -cksum;
  if (cksum == b) return 1;            // Test checksum
  return 0;
}

/**********************************************************************/

void loadhexfile(FILE *f) {            // Loads a hex file into memory arrays
  char s[256];
  word i,lo,hi;
  word linetype = 0;   // 0 for data, 1 for end of file
  word wordcount;      // number of 16 bit words on this line
  word address;        // address where they begin
  word data[8];        // 16 bytes = 8 words max. per line of hex

  cleararrays();

  while((!feof(f)) && (linetype != 1)) {
    fgets(s,255,f);
    cleanctrl(s);

    if (!validhexline(s)) {            // Syntax check
      s[40] = 0;                       // Truncate invalid line for display
      if (s[0] != ':') {
	printf("  Invalid line (skipped): '%s'...\n",s);
        continue;
      }
      else {
	printf("  Unable to decode line:  '%s'...\n",s);
        goto bailout;
      }
    }

    sscanf(s+1,"%2x",&wordcount);      // Parse the line - Intel Hex8M
    wordcount = wordcount/2;           // (double bytes, addresses doubled)
    sscanf(s+3,"%4x",&address);
    address = address/2;
    sscanf(s+7,"%2x",&linetype);

    if (linetype==1) goto finished;

    for (i=0; i<wordcount; i++) {           // Grab the data
      sscanf(s+(9+4*i),"%2x%2x",&lo,&hi);
      data[i] = (hi << 8) | lo;
    }

                                            // Place in appropriate array
    if (address >= DBASE) {
      if (!stuffarray(address,DMEM,DBASE,DSIZE,&DUSED,data,wordcount))
        goto bailout;
    }
    else if (address >= CBASE) {
      if (!stuffarray(address,CMEM,CBASE,CSIZE,&CUSED,data,wordcount))
        goto bailout;
    }
    else if (address >= IBASE) {
      if (!stuffarray(address,IMEM,IBASE,ISIZE,&IUSED,data,wordcount))
        goto bailout;
    }
    else {
      if (!stuffarray(address,PMEM,PBASE,PSIZE,&PUSED,data,wordcount))
        goto bailout;
    }
  } // while

finished:
  printf("  Program memory loaded: %5d word(s)\n",PUSED);
  printf("  Configuration loaded:  %5d word(s)\n",CUSED);
  printf("  ID memory loaded:      %5d word(s)\n",IUSED);
  printf("  Data memory loaded:    %5d byte(s)\n",DUSED);
  return;

bailout:
  cleararrays();
  errmsg("  Unable to load file.");
  FNAME[0] = 0;
  return;
}

// *********************************************************************
// USER INTERFACE
// *********************************************************************

void banner() {

  clearscreen();

  puts("  -----------------------------------");

#ifdef DOS
  puts("  NOPPP - \"No-Parts\" PIC Programmer");
  puts("  Michael A. Covington");
  puts("  Version of " __DATE__ " " __TIME__);
#endif

#ifdef LINUX
  puts("   Linux PIC Programmer");
  puts("   P. J. Rice - " __DATE__ " " __TIME__);
#endif

  puts("  -----------------------------------");
  if (LPT > 0) printf("   Using LPT%d on %03XH\n",LPT,PORT);
  puts("  -----------------------------------");
  printf("   %s    %s\n", pic_name, FNAME);
  puts("  -----------------------------------\n");
}

/**********************************************************************/

void selectport() {

    LPT = 3;
    PORT = portaddr(LPT);
    if( !detecthardware() ) {

	LPT = 2;
	PORT = portaddr(LPT);
	if( !detecthardware() ) {

	    LPT = 1;
	    PORT = portaddr(LPT);
	    if( !detecthardware() ) {

		CLAIM_PORT
		
		banner();
		puts("  Caution: Programmer hardware not found!\n\n");
		puts("  Defaulting to LPT1");
		puts("  Press Ctrl-C to cancel or");
		errmsg("   ");

	    }
	}
    }
}

/**********************************************************************/

int troubleshoot() {

  allpinslow();
  banner();
  puts("  TEST A\n");
  puts("  Connect negative voltmeter lead to pin 5");
  puts("  of PIC socket and check the following voltages:\n");
  puts("  Ensure no PIC is in the socket.");
  puts("    Socket pin  4      < 0.8 V");
  puts("    Socket pin 12      < 0.8 V");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      < 0.8 V");
  puts("    Green  LED should be ON            <<");
  puts("    Red    LED should be OFF");
  puts("    Orange LED should be OFF");
  errmsg("   ");

  VCCON
  banner();
  puts("  TEST B\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      < 0.8 V");
  puts("    Socket pin 12      < 0.8 V");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      4.75 to 5.25 V  <<");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON            <<");
  puts("    Orange LED should be OFF");
  errmsg("  ");

  vppon();
  banner();
  puts("  TEST C\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      12.0 to 14.0 V   <<");
  puts("    Socket pin 12      < 0.8 V");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be ON            <<");
  errmsg("   ");

  CLOCKUP
  banner();
  puts("  TEST D\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      12.0 to 14.0 V");
  puts("    Socket pin 12      4.75 to 5.25 V  <<");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be ON");
  errmsg("  ");

  DATAUP
  banner();
  puts("  TEST E\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      12.0 to 14.0 V");
  puts("    Socket pin 12      4.75 to 5.25 V");
  puts("    Socket pin 13      4.75 to 5.25 V  <<");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be ON");
  errmsg("  ");

  DATADOWN
  banner();
  puts("  TEST F\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      12.0 to 14.0 V");
  puts("    Socket pin 12      4.75 to 5.25 V");
  puts("    Socket pin 13      < 0.8 V         <<");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be ON");
  errmsg("  ");

  CLOCKDOWN
  banner();
  puts("  TEST G\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      12.0 to 14.0 V");
  puts("    Socket pin 12      < 0.8 V         <<");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be ON");
  errmsg("  ");

  VPPOFF
  banner();
  puts("  TEST H\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      < 0.8 V         <<");
  puts("    Socket pin 12      < 0.8 V");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      4.75 to 5.25 V");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be ON");
  puts("    Orange LED should be OFF           <<");
  errmsg("   ");

  VCCOFF
  banner();
  puts("  TEST I\n");
  puts("  With negative voltmeter lead still");
  puts("  connected to pin 5 in the PIC socket,");
  puts("  check the following voltages:\n");
  puts("    Socket pin  4      < 0.8 V");
  puts("    Socket pin 12      < 0.8 V");
  puts("    Socket pin 13      < 0.8 V");
  puts("    Socket pin 14      < 0.8 V         <<");
  puts("    Green  LED should be ON");
  puts("    Red    LED should be OFF           <<");
  puts("    Orange LED should be OFF");
  errmsg("  ");

  banner();
  puts("  This completes the voltage test sequence.\n");
  errmsg("   ");

  allpinslow();

#ifdef LINUX
  system("stty -cbreak echo");
#endif

  exit(0);
}

/**********************************************************************/

void inspect() {

	word i,w;
	banner();

	//  VCCON
	vppreset();
  
	if( DEVICE <= PIC12C509A ) {  // Read Config word first
		HIGH_P
		sendcmd(READPROGRAM);
		w = (recvdata() & 0x0FFFF);
		printf("  Configuration Word\n  Address 0FFF: PIC = %04X\n", w);
		sendcmd(INCREMENTADDRESS);
		LOW_P
		waitkey();
	}

	puts("  Program memory:");
	if( !readall(PMASK,READPROGRAM,PMEM,PBASE,PSIZE))
		return;
  
        HIGH_P
	sendcmd(LOADCONFIG);      // from here on we're in config/ID memory
	senddata(DEFAULTCONFIG);  // loadconfig requires an arg, here it is
	LOW_P
  
	waitkey();

	puts("  Identification memory:");
	if( !readall(IMASK,READPROGRAM,IMEM,IBASE,ISIZEMAX))
		return;
  
	if( DEVICE > PIC12C509A )  {
	 
		  HIGH_P
		  for (i=0; i < CBASE-IBASE-ISIZEMAX; i++)
			sendcmd(INCREMENTADDRESS);      // get to config memory
		  LOW_P

		  puts("\n  Configuration memory:");
		  if( !readall(CMASK,READPROGRAM,CMEM,CBASE,CSIZEMAX))
			return;

		  vppreset();   // Reset address counter in PIC to 0

		  waitkey();
		  puts("  Data memory:");
		  if( !readall(DMASK,READDATA,DMEM,DBASE,DSIZEMAX))
			return;
	}
  
	puts("\n");

finish:
	allpinslow();
	waitkey();
}

bit readall(
// Read PIC memory.
// Returns true.
             word mask,                // to throw away irrelevant bits
             byte readcommand,         // command to read this memory,
             word array[],             // which memory array we're using
             word base,                // its base address
             word used) {              // number of valid words in array
  word i,w;

  int crlf = 0;
  
  for (i=0; i<used; i++) {
    HIGH_P
    sendcmd(readcommand);
    w = (recvdata() & mask);
    sendcmd(INCREMENTADDRESS);
    LOW_P
    printf("  Address %04X: PIC = %04X Prog = %04X\n", i+base, w, (array[i] & mask));
    crlf++;

    if( !(crlf%20) ) {
      crlf = 0;
      if( waitkey() != 0x20)
        return 0;
    }
  }
  return 1;
}

/**********************************************************************/

void load() {
  FILE *f;
  allpinslow();
  banner();
  printf("  File to load: ");
  strcpy(FNAME,getstring());
  f = fopen(FNAME,"rt");
  if (f == NULL) {
    errmsg("  Unable to open file.");
    FNAME[0] = 0;
    return;
  }
  loadhexfile(f);
  fclose(f);
  errmsg("  Loading complete.");

  if (CUSED == 0) {
    banner();
    puts("  Caution: HEX file did not contain a configuration word.\n");
    puts("  The following settings will be used:\n");
    puts("    RC oscillator");
    puts("    Watchdog timer disabled");
    puts("    Power-up timer enabled");
    puts("    Code not read-protected\n");
    errmsg("  You can specify other settings in the assembler.");
  }
  else if (CMEM[0] != (CMEM[0] & CMASK)) {
    banner();
    puts("  Caution: Configuration word appears to contain invalid bits.\n");
    puts("  Your program may have been assembled for a different");
    puts("  type of PIC.  Check device selection carefully.");
    errmsg("  ");
  }

}

/**********************************************************************/

void selectdevice() {

	int	i;

	allpinslow();
	pic_name = "         ";
	FNAME[0] = 0;	/* Declare "No File Loaded" */
	banner();

	// Clean out arrays

	for(i=0; i<PSIZEMAX; i++ ) PMEM[i] = 0x3fff;
	for(i=0; i<ISIZEMAX; i++ ) IMEM[i] = 0x3fff;
	for(i=0; i<CSIZEMAX; i++ ) CMEM[i] = 0x3fff;
	for(i=0; i<DSIZEMAX; i++ ) DMEM[i] = 0x3fff;

	PUSED = 0;
	CUSED = 0;
	IUSED = 0;
	DUSED = 0;

	puts("  Devices supported:\n");
	puts("    1   PIC12C508    2   PIC12C508A");
	puts("    3   PIC12C509    4   PIC12C509A");
	puts("    5   PIC16C83     6   PIC16C84");
	puts("    7   PIC16F83     8   PIC16F84");
	puts("    9   PIC16F627    A   PIC16F628");
	puts("    B   PIC16F648    C   PIC12F675\n");
	puts("    T   Test the programmer circuit\n");

	getchoice("1,2,3,4,5,6,7,8,9,A,B,C,T");

  switch(CHOICE) {

	case '1':	DEVICE        =  PIC12C508;
			pic_name      = "PIC12C508";
			erase_p       = erase;
			program_p     = program_12c5;
			PSIZE         =  512;
			DEFAULTCONFIG = 0x000a;
			Vpptime       = LATE;
			VPPlevel      = HIGH;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x0200; // of each memory
			CBASE         = 0x0FFF;
			DBASE         = 0x2100; // Unused
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;
			
	case '2':	DEVICE        =  PIC12C508A;
			pic_name      = "PIC12C508A";
			erase_p       = erase;
			program_p     = program_12c5;
			PSIZE         =  511;
			DEFAULTCONFIG = 0x000a;
			Vpptime       = LATE;
			VPPlevel      = HIGH;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x0200; // of each memory
			CBASE         = 0x0FFF;
			DBASE         = 0x2100; // Unused
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;
			
	case '3':	DEVICE        =  PIC12C509;
			pic_name      = "PIC12C509";
			erase_p       = erase;
			program_p     = program_12c5;
			PSIZE         = 1024;
			DEFAULTCONFIG = 0x000a;
			Vpptime       = LATE;
			VPPlevel      = HIGH;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x0200; // of each memory
			CBASE         = 0x0FFF;
			DBASE         = 0x2100; // Unused
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case '4':	DEVICE        =  PIC12C509A;
			pic_name      = "PIC12C509A";
			erase_p       = erase;
			program_p     = program_12c5;
			PSIZE         = 1023;
			DEFAULTCONFIG = 0x000a;
			Vpptime       = LATE;
			VPPlevel      = HIGH;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x0200; // of each memory
			CBASE         = 0x0FFF;
			DBASE         = 0x2100; // Unused
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case '5':	DEVICE        =  PIC16C83;
			pic_name      = "PIC16C83";
			erase_p       = erase;
			program_p     = program;
			PSIZE         =  512;
			DEFAULTCONFIG = 0x001B;
			Vpptime       = LATE;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;
			
	case '6':	DEVICE        =  PIC16C84;
			pic_name      = "PIC16C84";
			erase_p       = erase;
			program_p     = program;
			PSIZE         = 1024;
			DEFAULTCONFIG = 0x001B;
			Vpptime       = LATE;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x001f;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case '7':	DEVICE        =  PIC16F83;
			pic_name      = "PIC16F83";
			erase_p       = erase;
			program_p     = program;
			PSIZE         =  512;
			DEFAULTCONFIG = 0x3FF3;
			Vpptime       = LATE;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;
			
	case '8':	DEVICE        =  PIC16F84;
			pic_name      = "PIC16F84";
			erase_p       = erase;
			program_p     = program;
			PSIZE         = 1024;
			DEFAULTCONFIG = 0x3FF3;
			Vpptime       = LATE;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case '9':	DEVICE        =  PIC16F627;
			pic_name      = "PIC16F627";
			erase_p       =   erase628;
			program_p     = program;
			PSIZE         = 1024;
			DEFAULTCONFIG = 0x3F50;
			Vpptime       = EARLY;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case 'A':	DEVICE        =  PIC16F628;
			pic_name      = "PIC16F628";
			erase_p       =   erase;	// Was erase628()
			program_p     = program;
			PSIZE         = 2048;
			DEFAULTCONFIG = 0x3F50;
			Vpptime       = EARLY;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;
			
	case 'B':	DEVICE        =  PIC16F648;
			pic_name      = "PIC16F648";
			erase_p       =   erase628;
			program_p     = program;
			PSIZE         = 4096;
			DEFAULTCONFIG = 0x3F50;
			Vpptime       = EARLY;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case 'C':	DEVICE        =  PIC12F675;
			pic_name      = "PIC12F675";
			erase_p       =   erase675;
			program_p     = program675;
			PSIZE         = 1024;
			DEFAULTCONFIG = 0x3FF3;
			Vpptime       = EARLY;
			VPPlevel      = LOW;
			PBASE         = 0x0000; // Base address
			IBASE         = 0x2000; // of each memory
			CBASE         = 0x2007;
			DBASE         = 0x2100;
			PMASK         = 0x3fff;
			CMASK         = 0x3FFF;
			IMASK         = 0x3fff;
			DMASK         = 0x00ff;
			break;

	case 'T':	troubleshoot();
  }
}

/**********************************************************************/

void erase() {

//  PIC16F84

    int i;

//  VCCON
    vppreset();

    printf("  Commanding PIC to erase ID, configuration, ");

    HIGH_P
    sendcmd(LOADCONFIG);
    senddata(0x3FFF);

    for (i=7; i>0; i--)
        sendcmd(INCREMENTADDRESS);

    sendcmd(1);
    sendcmd(7);
    sendcmd(BEGINPROGRAMMING);
    MyDelay(20000);
    sendcmd(1);
    sendcmd(7);
    LOW_P

    printf("  program, ");
    HIGH_P
    progcycle(ERASEPROGRAM,0x3FFF);    // is the data word necessary?
    LOW_P

    printf("  data...\n");
    HIGH_P
    progcycle(ERASEDATA,0x3FFF);       // is the data word necessary?
    LOW_P

    allpinslow();

    puts("   Done.");
    waitkey();
}

/**********************************************************************/

void erase675() {

//	Erase PIC12F675

	read675cal();
	
//	Keep only bandgap bits, set all other settable bits.
	
	bandgap = (bandgap & 0x3000) | 0x1FF;

//	VCCON
	vppreset();


	printf("  Erasing PIC\n");
  
	HIGH_P
  
	sendcmd(LOADCONFIG);	// Section 4.0 of F675
	senddata(0x3FFF);	// programming notes
	MyDelay(20000);		// T erase >= 8ms
  
	sendcmd(ERASEPROGRAM);	// No data reqd.
	MyDelay(20000);		// T erase >= 8ms
  
	sendcmd(ERASEDATA);	// No data reqd.
	MyDelay(20000);		// T erase >= 8ms
  
	LOW_P

	allpinslow();
	write675cal();
	allpinslow();
	puts("  Done.");
	waitkey();
}

/**********************************************************************/

void erase628() {

//	Erase PIC16F627A/628A/648A
/*	Appears not to work, whereas the 16F84 erase() function
	does work. So we're using that for the time being */
	
	vppreset();

	printf("  Erasing PIC\n");
	
	HIGH_P

	sendcmd(LOADCONFIG);
	senddata(0x3FFF);
	MyDelay(20000);		// T erase >= 6ms
 
	sendcmd(LOADPROGRAM);	// Section 3.1 of 16F628 Prog Spec
	senddata(0x3FFF);
	sendcmd(ERASEPROGRAM);
	MyDelay(20000);		// T erase >= 6ms

	sendcmd(ERASEDATA);
	MyDelay(20000);		// T erase >= 6ms

	LOW_P

	allpinslow();

	puts("  Done.");
	waitkey();
}

/**********************************************************************/

void repair675() {

// Only used for 12F675 to manually fix buggered calibration

#ifdef LINUX
	system("stty -cbreak echo");
//	fflush(stdin);
#endif

  printf("\n  Enter Oscillator Calibration Word: ");
 
  if( scanf("%x", &osccal) != 1) {
      printf("\n  Couldn't make sense of input\n");
      goto finish;
  }
  
  osccal &= PMASK;
  
  printf("  Enter Bandgap Calibration Word: ");
  
  if( scanf("%x", &bandgap) != 1) {
      printf("\n  Couldn't make sense of input\n");
      goto finish;
  }

  bandgap &= CMASK;

  vppreset();

  printf("  Erasing PIC\n");
  
	HIGH_P
  
	sendcmd(LOADCONFIG);	// Section 4.0 of F675
	senddata(0x3FFF);	// programming notes
	MyDelay(20000);		// T erase >= 8ms
  
	sendcmd(ERASEPROGRAM);	// No data reqd.
	MyDelay(20000);		// T erase >= 8ms
  
	sendcmd(ERASEDATA);	// No data reqd.
	MyDelay(20000);		// T erase >= 8ms
  
	LOW_P

  allpinslow();

  write675cal();

  puts(" ");

finish:

#ifdef LINUX
	system("stty cbreak -echo");
#endif

  allpinslow();

  puts("\n  Press space bar to continue...");

  while (getchar() != 0x20 ) {};
}

/**********************************************************************/

void read675cal() {

  word i,w;

//  VCCON
  vppreset();

  printf("  Reading configuration.\n");

  HIGH_P
  for (i=0; i<(PSIZE-1); i++)        // Skip the program, goto last location    
    sendcmd(INCREMENTADDRESS);

  sendcmd(READPROGRAM);
  osccal = (recvdata() & PMASK);     // Got the OSC Cal instruction RETLW xx
  LOW_P
  
  if( (osccal & 0x3C00) != 0x3400 )
      printf("  Invalid Oscillator Calibration = %04x\n", osccal );
  else
      printf("  Valid Oscillator Calibration   = %04x\n", osccal );

  HIGH_P
  sendcmd(LOADCONFIG);               // from here on we're in config/ID memory
  senddata(DEFAULTCONFIG);           // loadconfig requires an arg, here it is

  for (i=0; i<7; i++)                // Skip the ID, goto last location
       sendcmd(INCREMENTADDRESS);
  
  sendcmd(READPROGRAM);
  bandgap = (recvdata() & CMASK);   // Band-gap bits
  LOW_P
  
  printf("  Band-Gap Calibration           = %04x\n", bandgap );
  allpinslow();
}

/**********************************************************************/

void write675cal() {

  word i,w;

  printf("  Writing configuration.\n");

//  VCCON
  vppreset();

  HIGH_P
  for (i=0; i<(PSIZE-1); i++)        // Skip the program, goto last location
     sendcmd(INCREMENTADDRESS);
	
  progcycle(LOADPROGRAM,osccal);    // Write osccal

  sendcmd(READPROGRAM);
  w = (recvdata() & PMASK);
  LOW_P

    if (w != osccal)
      printf("  Failed to write osccal: Expecting %04X, found %04X.\n", osccal, w);

  HIGH_P
  sendcmd(LOADCONFIG);               // from here on we're in config/ID memory
  senddata(DEFAULTCONFIG);           // loadconfig requires an arg, here it is

  for (i=0; i<7; i++)                // Skip the ID, goto last location
     sendcmd(INCREMENTADDRESS);

  progcycle(LOADPROGRAM, bandgap);  // Write bandgap
  sendcmd(READPROGRAM);
  w = (recvdata() & CMASK);
  LOW_P
  
  if (w != bandgap )
     printf("  Failed to write bandgap: Expecting %04X, found %04X.\n", bandgap, w);

  allpinslow();
}

/**********************************************************************/

void program(int mode) {

//	PIC16F84

  word i;

  banner();

  if (PUSED+IUSED+CUSED+DUSED == 0) {
    printf("  Load a file first.\n");
    goto finish;
  }

//  VCCON
  vppreset();

  printf("  Program memory:       ");

  if (!programall(mode,PMASK,LOADPROGRAM,READPROGRAM,PMEM,PBASE,PUSED))
    goto finish;

  HIGH_P
  sendcmd(LOADCONFIG);      // from here on we're in config/ID memory
  senddata(DEFAULTCONFIG);  // loadconfig requires an arg, here it is
  LOW_P

  printf("  ID memory:            ");
  if (!programall(mode,IMASK,LOADPROGRAM,READPROGRAM,IMEM,IBASE,IUSED))
    goto finish;

  HIGH_P
  for (i=0; i < CBASE-IBASE-IUSED; i++)
    sendcmd(INCREMENTADDRESS);      // get to config memory
  LOW_P

  printf("  Configuration memory: ");
  if (!programall(mode,CMASK,LOADPROGRAM,READPROGRAM,CMEM,CBASE,CUSED))
    goto finish;

  vppreset();   // Reset address counter in PIC to 0

  printf("  Data memory:          ");
  if (!programall(mode,DMASK,LOADDATA,READDATA,DMEM,DBASE,DUSED))
    goto finish;

  puts("\n  Programming complete.\n\n");
  puts("  For production-grade work, you should now verify");
  puts("  the PIC at the maximum and minimum values of Vcc.");

finish:
  allpinslow();
  waitkey();
}

/**********************************************************************/

void program675(int mode) {

//	PIC12F675

  word i;

  banner();

  if (PUSED+IUSED+CUSED+DUSED == 0) {
    printf("  Load a file first.\n");
    goto finish;
  }

  if( mode != VERIFY ) {

       read675cal();

       if( PMEM[PSIZE-1] != osccal)
	   printf("  Updating osccal: was %04x   now %04x\n", PMEM[PSIZE-1] , osccal);

       PMEM[PSIZE-1] = osccal;
  //     PUSED = PSIZE;

       if( (CMEM[0] & 0x3000) != (bandgap & 0x3000) )
	   printf("  Updating config: was %04x   now %04x\n", CMEM[0] , (CMEM[0] & 0x01FF) | (bandgap & 0x3000) );

       CMEM[0] = (CMEM[0] & 0x01FF) | (bandgap & 0x3000);
  }
  
//  VCCON
  vppreset();

  printf("  Program memory:        ");

  if (!programall(mode,PMASK,LOADPROGRAM,READPROGRAM,PMEM,PBASE,PUSED))
    goto finish;

  if( PUSED < PSIZE ) {  // then we have to write the osc cal word too
  
      HIGH_P
      for (i=0; i < PSIZE-PUSED; i++)
	sendcmd(INCREMENTADDRESS);      // get to osc cal word
      LOW_P

      printf("  Oscillator calibration ");

      if (!programall(mode,PMASK,LOADPROGRAM,READPROGRAM,PMEM,PSIZE-1,1))
	goto finish;
  }

  HIGH_P
  sendcmd(LOADCONFIG);      // from here on we're in config/ID memory
  senddata(DEFAULTCONFIG);  // loadconfig requires an arg, here it is
  LOW_P

  printf("  ID memory:             ");
  if (!programall(mode,IMASK,LOADPROGRAM,READPROGRAM,IMEM,IBASE,IUSED))
    goto finish;

  HIGH_P
  for (i=0; i < CBASE-IBASE-IUSED; i++)
    sendcmd(INCREMENTADDRESS);      // get to config memory
  LOW_P

  printf("  Configuration memory:  ");
  if (!programall(mode,CMASK,LOADPROGRAM,READPROGRAM,CMEM,CBASE,CUSED))
    goto finish;

  vppreset();   // Reset address counter in PIC to 0

  printf("  Data memory:           ");
  if (!programall(mode,DMASK,LOADDATA,READDATA,DMEM,DBASE,DUSED))
    goto finish;

  puts("\n  Programming complete.\n\n");
  puts("  For production-grade work, you should now verify");
  puts("  the PIC at the maximum and minimum values of Vcc.");

finish:
  allpinslow();
  waitkey();
}

/**********************************************************************/

void program_12c5( int mode ) {

	word	data, readback;
	int     i, times;

	banner();
	if ( PUSED+IUSED+CUSED+DUSED == 0) {
		printf("  Load a file first.\n");
		goto finish;
	}

	HIGH_P
	VCCON			// See note, bottom of Page 5 of 12C5 serial programming notes
	MyDelay(100);
	vppon();
	MyDelay(100);

	sendcmd(INCREMENTADDRESS);	// Skip config word, we do it last.
	MyDelay(1);
	LOW_P

	printf("  Program memory:       ");

	if (!programall_12c5(mode,PMASK,LOADPROGRAM,READPROGRAM,PMEM,PBASE,PUSED))
		goto finish;

	HIGH_P
	sendcmd(LOADCONFIG);      // from here on we're in config/ID memory
	senddata(DEFAULTCONFIG);  // loadconfig requires an arg, here it is
	LOW_P

	printf("  ID memory:            ");
	if (!programall_12c5(mode,IMASK,LOADPROGRAM,READPROGRAM,IMEM,IBASE,IUSED))
		goto finish;

	HIGH_P
	VPPOFF			// Reset address counter in PIC to 0
	MyDelay(200);
	DATADOWN
	CLOCKDOWN
	VCCOFF

	MyDelay(25000);

	VCCON			// See note, bottom of Page 5 of 12C5 serial programming notes
	MyDelay(100);
	vppon();
	MyDelay(100);
	LOW_P

	printf("  Configuration memory: ");
	if (!programall_12c5(mode,CMASK,LOADPROGRAM,READPROGRAM,CMEM,CBASE,CUSED))
		goto finish;

	puts("\n  Programming complete.\n\n");
	puts("  For production-grade work, you should now verify");
	puts("  the PIC at the maximum and minimum values of Vcc.");

finish:
	allpinslow();
	waitkey();
}

bit programall_12c5(
// Program a memory from appropriate array.
// Returns true if successful.
             int  mode,                // program or verify
             word mask,                // to throw away irrelevant bits
             byte writecommand,        // command to write this memory
             byte readcommand,         // command to read this memory,
             word array[],             // which memory array we're using
             word base,                // its base address
             word used) {              // number of valid words in array

	word i,w;

	switch(mode) {
		case PROGRAM: puts("  programming..."); break;
		case VERIFY:  puts("  verifying..."); break;
	}

	for (i=0; i<used; i++) {

		printf("  %04X\b\b\b\b\b\b",i+base);     // display addr and backspace

		HIGH_P
		if (mode==PROGRAM)
			if(!progcycle_12c5( array[i] & mask, mask, base )) {
				LOW_P
				printf("  Failed at %04X:         Count exceeded!\n", i+base);
				return 0;
			}

		sendcmd(readcommand);

		w = (recvdata() & mask);
		if (w != (array[i] & mask)) {
			LOW_P
			printf("  Failed at %04X:         Expecting %04X, found %04X.\n",
	                                         i+base, (array[i] & mask), w);
			return 0;
		}
		sendcmd(INCREMENTADDRESS);
		LOW_P
	}
	return 1;
}


int progcycle_12c5( word arg, word mask, word base ) {	// Do one word

int     i, times;
word    readback;

	if( base == 0x0FFF ) {

		// We are doing the config word.
		// Don't verify yet, just blast away for 10ms total!
		
		for( times=0; times <100; times++)
				tiny_p_cycle( arg );

		readback = (recvdata() & mask);

		if( readback == arg)
			return 1;	// & return OK status
		else
			return 0;	// Failed on config word
	
	} else {

		// Doing a "normal" EPROM location

		if( arg != mask ) {	// Do only if it needs it
			
			times = 0;

			do {	tiny_p_cycle( arg & mask );
				times++;
				sendcmd(READPROGRAM);
				MyDelay(1);
				readback = (recvdata() & mask);

			} while( (times <= 8) && ( readback != arg ) );

			if( times > 8 )
				return 0;	// We have buggered it!

			for( i=0; i< (times*11); i++ )
				tiny_p_cycle( arg );  // OK so overprogram

		}
		// Final check that this location is OK
		readback = (recvdata() & mask);

		if( readback == arg)
			return 1;	// & return OK status
		else
			return 0;	// Failed on config word
	}	
}

void	tiny_p_cycle( word arg) {  // One cycle @ 100us per cycle

	sendcmd(LOADPROGRAM);
	MyDelay(1);
	senddata(arg);
	MyDelay(1);
	sendcmd(BEGINPROGRAMMING);
	MyDelay(100);
	sendcmd(ENDPROGRAMMING);
	MyDelay(1);
}

/**********************************************************************/

void queryexit() {
  allpinslow();
  banner();
  puts("  Are you sure you want to exit?");
  getchoice("Y,N");
  if (CHOICE == 'Y') {
    banner();
    errmsg("  You may remove the PIC from the socket now.");

#ifdef LINUX
  system("stty -cbreak echo");
#endif

    exit(0);
  }
}


/**********************************************************************/

void menu() {
  allpinslow(); // Clean up after aberrant routines if any
  banner();
  puts("    L  Load HEX file");
  puts("    S  Select type of PIC");

  if( DEVICE > PIC12C509A ) puts("    E  Erase PIC");
  
  if( DEVICE == PIC12F675 ) puts("    R  Erase + Repair Osc Cal & Bandgap");

  puts("    P  Program PIC");
  puts("    V  Verify PIC");
  puts("    I  Inspect PIC memory\n");
  puts("    X  Exit program");

  if( DEVICE == PIC12F675 )
      getchoice("L,S,E,R,P,V,I,X");	// Include Osc repair function

  else if( DEVICE > PIC12C509A )
      getchoice("L,S,E,P,V,I,X");

  else
      getchoice("L,S,P,V,I,X");		// Can't erase little buggers

  switch(CHOICE) {
    case 'L': load();             break;
    case 'S': selectdevice();     break;
    case 'E': erase_p();          break;
    case 'R': repair675();        break;
    case 'P': program_p(PROGRAM); break;
    case 'V': program_p(VERIFY);  break;
    case 'I': inspect();          break;
    case 'X': queryexit();
  }
}

/**********************************************************************/

main() {

#ifdef LINUX

	if( mlockall(MCL_CURRENT  | MCL_FUTURE) ) {
		printf("Cannot lock memory!\n\n");
		exit(1);
	}

	system("stty cbreak -echo");
#endif

	selectport();
	allpinslow();
//	FNAME[0] = 0;    // no file is presently loaded
	selectdevice();  // mandatory
	banner();
	puts("  You may insert the PIC in the socket now.");
	errmsg("  Be very careful not to insert it backward.");

	while(1) menu();
	allpinslow();
	return 0;
}
